import { objectLoop } from '@dzeio/object-util'
import { buildRFC7807 } from './RFCs/RFC7807'
import StatusCode from './StatusCode'

interface StorageItem {
	pointsRemaining: number
	timeReset: number
}

export interface RateLimitHeaders {
	"Retry-After"?: string,
	"X-RateLimit-Limit": string,
	"X-RateLimit-Remaining": string,
	"X-RateLimit-Reset": string
}

export default class RateLimiter {

	/**
	 * number of points that can be used per {timeSpan}
	 */
	public static points = 10

	/**
	 * timeSpan in seconds
	 */
	public static timeSpan = 600

	private static instance: RateLimiter = new RateLimiter()
	public static getInstance(): RateLimiter {
		return this.instance
	}

	private storage: Record<string, StorageItem> = {}

	public consume(key: string, value: number = 1): Response | RateLimitHeaders {
		let item = this.storage[key]
		const now = (new Date().getTime() / 1000)
		if (!item) {
			item = {
				pointsRemaining: RateLimiter.points,
				timeReset: now + RateLimiter.timeSpan
			}
		}
		if (item.timeReset <= now) {
			item.timeReset = now + RateLimiter.timeSpan
			item.pointsRemaining = RateLimiter.points
		}
		item.pointsRemaining -= value
		this.storage[key] = item
		const headers: RateLimitHeaders = {
			"X-RateLimit-Limit": RateLimiter.points.toFixed(0),
			"X-RateLimit-Remaining": Math.max(item.pointsRemaining, 0).toFixed(0),
			"X-RateLimit-Reset": item.timeReset.toFixed(0)
		}
		if (item.pointsRemaining < 0) {
			const resp = buildRFC7807({
				type: '/docs/error/rate-limited',
				status: StatusCode.TOO_MANY_REQUESTS,
				title: 'You are being rate limited as you have done too many requests to the server'
			})
			resp.headers.append('Retry-After', (item.timeReset - now).toFixed(0))
			objectLoop(headers, (value, key) => {
				resp.headers.append(key, value)
			})
			return resp
		}
		return headers
	}
}
